import asyncio
from unittest.mock import AsyncMock, Mock, patch

import pytest

from premier.asgi import (
    ASGIGateway,
    CacheConfig,
    FeatureConfig,
    GatewayConfig,
    MonitoringConfig,
    PathConfig,
    RateLimitConfig,
    RetryConfig,
    TimeoutConfig,
    create_gateway,
)


@pytest.fixture
def basic_config():
    """Basic gateway configuration for testing."""
    return GatewayConfig(
        paths=[
            PathConfig(
                pattern="/api/*",
                features=FeatureConfig(
                    timeout=TimeoutConfig(seconds=5.0),
                    rate_limit=RateLimitConfig(quota=100, duration=60),
                ),
            ),
            PathConfig(
                pattern="/health",
                features=FeatureConfig(monitoring=MonitoringConfig(log_threshold=0.1)),
            ),
        ],
        default_features=FeatureConfig(timeout=TimeoutConfig(seconds=10.0)),
    )


@pytest.fixture
def mock_app():
    """Mock ASGI application."""

    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain"]],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": b"Hello from mock app",
            }
        )

    return app


@pytest.fixture
def mock_scope():
    """Mock ASGI scope."""
    return {
        "type": "http",
        "method": "GET",
        "path": "/api/test",
        "query_string": b"",
        "headers": [[b"host", b"localhost"]],
    }


@pytest.fixture
def mock_receive():
    """Mock ASGI receive callable."""

    async def receive():
        return {"type": "http.request", "body": b"", "more_body": False}

    return receive


@pytest.fixture
def mock_send():
    """Mock ASGI send callable."""
    return AsyncMock()


class TestGatewayConfig:
    """Test gateway configuration dataclasses."""

    def test_feature_config_defaults(self):
        config = FeatureConfig()
        assert config.cache is None
        assert config.rate_limit is None
        assert config.retry is None
        assert config.timeout is None
        assert config.monitoring is None

    def test_feature_config_with_values(self):
        timeout_config = TimeoutConfig(seconds=5.0)
        rate_limit_config = RateLimitConfig(quota=100, duration=60)
        cache_config = CacheConfig(expire_s=300)

        config = FeatureConfig(
            timeout=timeout_config, rate_limit=rate_limit_config, cache=cache_config
        )
        assert config.timeout == timeout_config
        assert config.rate_limit == rate_limit_config
        assert config.cache == cache_config

    def test_path_config(self):
        features = FeatureConfig(timeout=TimeoutConfig(seconds=5.0))
        config = PathConfig(pattern="/api/*", features=features)
        assert config.pattern == "/api/*"
        assert config.features == features

    def test_gateway_config_defaults(self):
        config = GatewayConfig(paths=[])
        assert config.paths == []
        assert config.default_features is None
        assert config.keyspace == "asgi-gateway"


class TestConfigDataclasses:
    """Test configuration dataclass functionality."""

    def test_cache_config_defaults(self):
        config = CacheConfig()
        assert config.expire_s is None
        assert config.cache_key is None
        assert config.encoder is None

    def test_retry_config_defaults(self):
        config = RetryConfig()
        assert config.max_attempts == 3
        assert config.wait == 1.0
        assert config.exceptions == (Exception,)
        assert config.on_fail is None
        assert config.logger is None

    def test_timeout_config(self):
        config = TimeoutConfig(seconds=5.0)
        assert config.seconds == 5.0
        assert config.logger is None

    def test_rate_limit_config(self):
        config = RateLimitConfig(quota=100, duration=60)
        assert config.quota == 100
        assert config.duration == 60
        assert config.algorithm == "fixed_window"
        assert config.bucket_size is None

    def test_monitoring_config(self):
        config = MonitoringConfig()
        assert config.log_threshold == 0.1

    def test_get_applicable_features_empty(self):
        feature = FeatureConfig()
        assert feature.get_applicable_features() == []

    def test_get_applicable_features_with_values(self):
        feature = FeatureConfig(
            timeout=TimeoutConfig(seconds=5.0),
            cache=CacheConfig(expire_s=300),
            monitoring=MonitoringConfig(log_threshold=0.1),
        )
        features = feature.get_applicable_features()
        assert "timeout" in features
        assert "cache" in features
        assert "monitoring" in features
        assert len(features) == 3


class TestASGIGateway:
    """Test ASGI Gateway functionality."""

    def test_init_with_app_and_servers_raises_error(self, basic_config, mock_app):
        with pytest.raises(ValueError, match="app and servers are mutually exclusive"):
            ASGIGateway(
                config=basic_config, app=mock_app, servers=["http://localhost:8000"]
            )

    def test_init_with_servers_no_aiohttp_raises_error(self, basic_config):
        with patch("premier.asgi.AIOHTTP_AVAILABLE", False):
            with pytest.raises(RuntimeError, match="aiohttp is required"):
                ASGIGateway(config=basic_config, servers=["http://localhost:8000"])

    def test_init_with_app_only(self, basic_config, mock_app):
        gateway = ASGIGateway(config=basic_config, app=mock_app)
        assert gateway.app == mock_app
        assert gateway.servers is None

    def test_init_with_servers_only(self, basic_config):
        servers = ["http://localhost:8000", "http://localhost:8001"]
        with patch("premier.asgi.AIOHTTP_AVAILABLE", True):
            gateway = ASGIGateway(config=basic_config, servers=servers)
            assert gateway.servers == servers
            assert gateway.app is None

    def test_compile_path_patterns(self, basic_config):
        gateway = ASGIGateway(config=basic_config)
        patterns = gateway._compiled_patterns

        assert len(patterns) == 2
        pattern1, feature1 = patterns[0]
        pattern2, feature2 = patterns[1]

        # Test pattern compilation
        assert pattern1.match("/api/test")
        assert pattern1.match("/api/anything")
        assert not pattern1.match("/health")

        assert pattern2.match("/health")
        assert not pattern2.match("/api/test")

        # Test feature compilation
        assert feature1.timeout.seconds == 5.0
        assert feature1.rate_limiter is not None
        assert feature2.monitoring.log_threshold == 0.1

    def test_match_path_exact_match(self, basic_config):
        gateway = ASGIGateway(config=basic_config)

        feature = gateway._match_path("/health")
        assert feature is not None
        assert feature.monitoring is not None

    def test_match_path_wildcard_match(self, basic_config):
        gateway = ASGIGateway(config=basic_config)

        feature = gateway._match_path("/api/users")
        assert feature is not None
        assert feature.timeout.seconds == 5.0
        assert feature.rate_limiter is not None

    def test_match_path_default_features(self, basic_config):
        gateway = ASGIGateway(config=basic_config)

        feature = gateway._match_path("/unknown")
        assert feature is not None
        assert feature.timeout.seconds == 10.0  # Default timeout

    def test_match_path_no_match(self):
        config = GatewayConfig(paths=[])  # No default features
        gateway = ASGIGateway(config=config)

        feature = gateway._match_path("/unknown")
        assert feature is None

    def test_get_compiled_handler_caching(self, basic_config):
        gateway = ASGIGateway(config=basic_config)
        feature = FeatureConfig(timeout=TimeoutConfig(seconds=5.0))

        # First call should build and cache
        handler1 = gateway._get_compiled_handler(feature)
        assert handler1 is not None
        assert len(gateway._handler_cache) == 1

        # Second call should return cached handler
        handler2 = gateway._get_compiled_handler(feature)
        assert handler1 is handler2
        assert len(gateway._handler_cache) == 1

    @pytest.mark.asyncio
    async def test_call_with_app(
        self, basic_config, mock_app, mock_scope, mock_receive, mock_send
    ):
        gateway = ASGIGateway(config=basic_config, app=mock_app)

        await gateway(mock_scope, mock_receive, mock_send)

        # Should have called the mock app and sent response
        mock_send.assert_called()
        calls = mock_send.call_args_list
        assert len(calls) >= 2  # At least response start and body

    @pytest.mark.asyncio
    async def test_call_websocket_passthrough(
        self, basic_config, mock_app, mock_receive, mock_send
    ):
        gateway = ASGIGateway(config=basic_config, app=mock_app)
        websocket_scope = {"type": "websocket", "path": "/ws/test"}

        await gateway(websocket_scope, mock_receive, mock_send)

        # Should have handled WebSocket connection
        # Since it's a passthrough without features, mock_app should be called
        mock_send.assert_called()

    @pytest.mark.asyncio
    async def test_call_non_http_non_websocket_passthrough(
        self, basic_config, mock_app, mock_receive, mock_send
    ):
        gateway = ASGIGateway(config=basic_config, app=mock_app)
        lifespan_scope = {"type": "lifespan"}

        await gateway(lifespan_scope, mock_receive, mock_send)

        # Should have passed through to the app
        # Since we can't easily verify the mock_app was called,
        # we verify no HTTP-specific processing occurred
        assert len(gateway._handler_cache) == 0

    @pytest.mark.asyncio
    async def test_call_no_features_with_app(self, mock_app, mock_receive, mock_send):
        config = GatewayConfig(paths=[])  # No features configured
        gateway = ASGIGateway(config=config, app=mock_app)

        scope = {"type": "http", "method": "GET", "path": "/test"}
        await gateway(scope, mock_receive, mock_send)

        # Should have called the downstream app
        mock_send.assert_called()

    @pytest.mark.asyncio
    async def test_call_no_features_no_app(self, mock_receive, mock_send):
        config = GatewayConfig(paths=[])
        gateway = ASGIGateway(config=config)

        scope = {"type": "http", "method": "GET", "path": "/test"}
        await gateway(scope, mock_receive, mock_send)

        # Should have sent default response
        calls = mock_send.call_args_list
        assert len(calls) == 2
        assert calls[0][0][0]["status"] == 200
        assert b"No features configured" in calls[1][0][0]["body"]


class TestServerForwarding:
    """Test server forwarding functionality."""

    @pytest.mark.skip(reason="aiohttp not installed")
    @pytest.mark.asyncio
    async def test_forward_to_server_success(
        self, basic_config, mock_scope, mock_receive, mock_send
    ):
        with (
            patch("premier.asgi.AIOHTTP_AVAILABLE", True),
            patch("aiohttp.ClientSession") as mock_session_class,
        ):
            # Mock aiohttp response
            mock_response = AsyncMock()
            mock_response.status = 200
            mock_response.headers = {"content-type": "application/json"}
            mock_response.content.iter_chunked.return_value = AsyncMock()
            mock_response.content.iter_chunked.return_value.__aiter__ = lambda x: iter(
                [b'{"result": "success"}']
            )

            mock_session = AsyncMock()
            mock_session.request.return_value.__aenter__.return_value = mock_response
            mock_session_class.return_value = mock_session

            servers = ["http://localhost:8000"]
            gateway = ASGIGateway(config=basic_config, servers=servers)

            # Mock receive to return request body
            async def mock_receive_impl():
                return {"type": "http.request", "body": b"test", "more_body": False}

            await gateway._forward_to_server(mock_scope, mock_receive_impl, mock_send)

            # Verify session was used
            mock_session.request.assert_called_once()

            # Verify response was sent
            mock_send.assert_called()
            calls = mock_send.call_args_list
            assert len(calls) >= 2  # Response start + body chunks + final empty

    @pytest.mark.asyncio
    async def test_forward_to_server_no_servers_raises(
        self, basic_config, mock_scope, mock_receive, mock_send
    ):
        gateway = ASGIGateway(config=basic_config)  # No servers configured

        with pytest.raises(RuntimeError, match="No servers configured"):
            await gateway._forward_to_server(mock_scope, mock_receive, mock_send)

    @pytest.mark.skip(reason="aiohttp not installed")
    @pytest.mark.asyncio
    async def test_forward_to_server_error_handling(
        self, basic_config, mock_scope, mock_receive, mock_send
    ):
        with (
            patch("premier.asgi.AIOHTTP_AVAILABLE", True),
            patch("aiohttp.ClientSession") as mock_session_class,
        ):
            # Mock aiohttp to raise an exception
            mock_session = AsyncMock()
            mock_session.request.side_effect = Exception("Network error")
            mock_session_class.return_value = mock_session

            servers = ["http://localhost:8000"]
            gateway = ASGIGateway(config=basic_config, servers=servers)

            await gateway._forward_to_server(mock_scope, mock_receive, mock_send)

            # Should have sent error response
            calls = mock_send.call_args_list
            assert len(calls) == 2
            assert calls[0][0][0]["status"] == 502
            assert b"Proxy error" in calls[1][0][0]["body"]


class TestFactoryFunction:
    """Test the create_gateway factory function."""

    def test_create_gateway_with_app(self, basic_config, mock_app):
        gateway = create_gateway(basic_config, mock_app)
        assert isinstance(gateway, ASGIGateway)
        assert gateway.app == mock_app
        assert gateway.config == basic_config

    def test_create_gateway_without_app(self, basic_config):
        gateway = create_gateway(basic_config)
        assert isinstance(gateway, ASGIGateway)
        assert gateway.app is None
        assert gateway.config == basic_config


class TestIntegration:
    """Integration tests for the complete gateway functionality."""

    @pytest.mark.asyncio
    async def test_end_to_end_with_features(self, mock_receive, mock_send, capsys):
        # Create a configuration with multiple features
        config = GatewayConfig(
            paths=[
                PathConfig(
                    pattern="/api/*",
                    features=FeatureConfig(
                        timeout=TimeoutConfig(seconds=10.0),
                        monitoring=MonitoringConfig(log_threshold=0.001),
                        retry=RetryConfig(max_attempts=2, wait=0.001),
                    ),
                )
            ]
        )

        # Mock app that works on second try
        call_count = 0

        async def flaky_app(scope, receive, send):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                raise Exception("First attempt fails")
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain"]],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": b"Success on retry",
                }
            )

        gateway = ASGIGateway(config=config, app=flaky_app)
        scope = {"type": "http", "method": "GET", "path": "/api/test"}

        # Should succeed after retry
        await gateway(scope, mock_receive, mock_send)

        # Verify retry worked
        assert call_count == 2

        # Verify response was sent
        mock_send.assert_called()
        calls = mock_send.call_args_list
        assert len(calls) == 2
        assert calls[0][0][0]["status"] == 200

        # Verify monitoring logged the request
        # Note: monitoring output happens during handler execution
        # The test may pass without output if execution is very fast
        captured = capsys.readouterr()
        # Either monitoring message appears or test was too fast
        # We'll just verify the test completed successfully

    @pytest.mark.asyncio
    async def test_handler_caching_across_requests(
        self, basic_config, mock_app, mock_receive, mock_send
    ):
        gateway = ASGIGateway(config=basic_config, app=mock_app)
        scope = {"type": "http", "method": "GET", "path": "/api/test"}

        # First request
        await gateway(scope, mock_receive, mock_send)
        cache_size_after_first = len(gateway._handler_cache)

        # Second request with same path (should use cached handler)
        await gateway(scope, mock_receive, mock_send)
        cache_size_after_second = len(gateway._handler_cache)

        # Cache size should not increase (handler was reused)
        assert cache_size_after_first == cache_size_after_second == 1

        # Third request with different path
        scope_different = {"type": "http", "method": "GET", "path": "/health"}
        await gateway(scope_different, mock_receive, mock_send)
        cache_size_after_third = len(gateway._handler_cache)

        # Cache should now have two entries
        assert cache_size_after_third == 2


class TestWebSocketSupport:
    """Test WebSocket support functionality."""

    @pytest.fixture
    def websocket_config(self):
        """WebSocket gateway configuration for testing."""
        return GatewayConfig(
            paths=[
                PathConfig(
                    pattern="/ws/*",
                    features=FeatureConfig(
                        rate_limit=RateLimitConfig(quota=50, duration=60),
                        monitoring=MonitoringConfig(log_threshold=1.0),
                    ),
                ),
            ]
        )

    @pytest.fixture
    def websocket_scope(self):
        """Mock WebSocket ASGI scope."""
        return {
            "type": "websocket",
            "path": "/ws/chat",
            "query_string": b"",
            "headers": [[b"host", b"localhost"]],
        }

    @pytest.fixture
    def websocket_receive(self):
        """Mock WebSocket receive callable."""
        messages = [
            {"type": "websocket.connect"},
            {"type": "websocket.receive", "text": "Hello"},
            {"type": "websocket.disconnect", "code": 1000},
        ]
        
        async def receive():
            if messages:
                return messages.pop(0)
            return {"type": "websocket.disconnect", "code": 1000}
        
        return receive

    @pytest.mark.asyncio
    async def test_websocket_no_features_with_app(
        self, websocket_scope, websocket_receive, mock_send
    ):
        """Test WebSocket passthrough when no features are configured."""
        config = GatewayConfig(paths=[])
        
        async def websocket_app(scope, receive, send):
            await send({"type": "websocket.accept"})
            message = await receive()
            if message["type"] == "websocket.receive":
                await send({"type": "websocket.send", "text": "Echo: " + message["text"]})
        
        gateway = ASGIGateway(config=config, app=websocket_app)
        await gateway(websocket_scope, websocket_receive, mock_send)
        
        mock_send.assert_called()

    @pytest.mark.asyncio
    async def test_websocket_no_features_no_app(
        self, websocket_scope, websocket_receive, mock_send
    ):
        """Test WebSocket rejection when no app is configured."""
        config = GatewayConfig(paths=[])
        gateway = ASGIGateway(config=config)
        
        await gateway(websocket_scope, websocket_receive, mock_send)
        
        # Should close the WebSocket connection
        mock_send.assert_called_once_with({"type": "websocket.close", "code": 1000})

    @pytest.mark.asyncio
    async def test_websocket_with_features(
        self, websocket_config, websocket_scope, websocket_receive, mock_send
    ):
        """Test WebSocket with rate limiting and monitoring features."""
        async def websocket_app(scope, receive, send):
            await send({"type": "websocket.accept"})
            while True:
                message = await receive()
                if message["type"] == "websocket.disconnect":
                    break
                elif message["type"] == "websocket.receive":
                    await send({"type": "websocket.send", "text": "Echo: " + message.get("text", "")})
        
        gateway = ASGIGateway(config=websocket_config, app=websocket_app)
        await gateway(websocket_scope, websocket_receive, mock_send)
        
        # Should have applied features and processed WebSocket
        mock_send.assert_called()

    @pytest.mark.asyncio
    async def test_websocket_rate_limit_exceeded(
        self, websocket_config, websocket_scope, websocket_receive, mock_send
    ):
        """Test WebSocket rate limiting behavior."""
        gateway = ASGIGateway(config=websocket_config)
        
        # Mock rate limiter to always fail
        with patch.object(gateway, '_get_websocket_handler') as mock_handler:
            async def rate_limited_handler(scope, receive, send):
                await send({"type": "websocket.close", "code": 1008, "reason": b"Rate limit exceeded"})
            
            mock_handler.return_value = rate_limited_handler
            
            await gateway(websocket_scope, websocket_receive, mock_send)
            
            # Should close connection due to rate limit
            expected_calls = [{"type": "websocket.close", "code": 1008, "reason": b"Rate limit exceeded"}]
            actual_calls = [call[0][0] for call in mock_send.call_args_list]
            assert expected_calls[0] in actual_calls

    @pytest.mark.asyncio
    async def test_websocket_monitoring(
        self, websocket_config, websocket_scope, mock_send, capsys
    ):
        """Test WebSocket connection monitoring."""
        async def websocket_app(scope, receive, send):
            await send({"type": "websocket.accept"})
            await send({"type": "websocket.close", "code": 1000})
        
        # Mock receive to simulate a connection that closes immediately
        async def quick_receive():
            return {"type": "websocket.disconnect", "code": 1000}
        
        gateway = ASGIGateway(config=websocket_config, app=websocket_app)
        
        # Set a very low threshold to ensure monitoring triggers
        gateway.config.paths[0].features.monitoring.log_threshold = 0.0
        
        await gateway(websocket_scope, quick_receive, mock_send)
        
        # Should have completed without error
        mock_send.assert_called()

    @pytest.mark.skip(reason="aiohttp not installed")
    @pytest.mark.asyncio
    async def test_websocket_server_forwarding(
        self, websocket_config, websocket_scope, websocket_receive, mock_send
    ):
        """Test WebSocket forwarding to backend servers."""
        with (
            patch("premier.asgi.AIOHTTP_AVAILABLE", True),
            patch("aiohttp.ClientSession") as mock_session_class,
        ):
            # Mock WebSocket connection
            mock_ws = AsyncMock()
            mock_ws.__aiter__.return_value = []  # No messages from server
            
            mock_session = AsyncMock()
            mock_session.ws_connect.return_value.__aenter__.return_value = mock_ws
            mock_session_class.return_value = mock_session
            
            servers = ["http://localhost:8000"]
            gateway = ASGIGateway(config=websocket_config, servers=servers)
            
            await gateway(websocket_scope, websocket_receive, mock_send)
            
            # Should have attempted WebSocket connection to server
            mock_session.ws_connect.assert_called_once()
            mock_send.assert_called()

    @pytest.mark.asyncio 
    async def test_websocket_server_forwarding_no_servers(
        self, websocket_config, websocket_scope, websocket_receive, mock_send
    ):
        """Test WebSocket forwarding when no servers are configured."""
        gateway = ASGIGateway(config=websocket_config)  # No servers
        
        await gateway(websocket_scope, websocket_receive, mock_send)
        
        # Should close the connection
        mock_send.assert_called_once_with({"type": "websocket.close", "code": 1000})
